#!/usr/bin/env python3

import numpy as np
import os
import sys

import wfdb
from utils import save_dict
import network
import tensorflow as tf

"""
Written by:  Muhammad Uzair Zahid

Save answers to '.json' files, the format is as {‘predict_endpoints’: [[s0, e0], [s1, e1], …, [sm-1, em-2]]}.
"""

def load_data(sample_path):
    sig, fields = wfdb.rdsamp(sample_path)
    length = len(sig)
    fs = fields['fs']

    return sig, length, fs

def prepare_windows(signal,win_size,stride,mean,std):

    """
    Prepare windows based on frame length and stride.
    """

    signal = np.squeeze(signal)

    # Lists of data windows
    data_windows = []

    # Split into windows and save corresponding padded indices
    for win_id in range(0, len(signal), stride):
        #print(win_id)
        if win_id + win_size < len(signal)+1:
            
            window = signal[win_id:win_id+win_size]
            data_windows.append(window)

    data_windows = np.asarray(data_windows)


    data_windows = (data_windows - mean)/std

    data_windows = np.expand_dims(data_windows,axis = 2)

    return data_windows

def get_label_idx(n_labels, stride = 10, win_size_label  = 20):
    """
    Prediction ids according to sliding windows. 
    """
    win_idx = []

    # Indices for padded signal
    id = np.arange(n_labels)

    # Split into windows and save corresponding padded indices
    for win_id in range(0, n_labels, stride):
        #print(win_id)
        if win_id + win_size_label < n_labels+1:
            
            win_idx.append(id[win_id:win_id+win_size_label])

    win_idx = np.asarray(win_idx)
    win_idx = win_idx.reshape(win_idx.shape[0]*win_idx.shape[1])

    return win_idx


def mean_preds(probs, win_idx):

    """
    Mean prediction of overlaping windows/frames.
    """

    probs_0 = probs[:,:,0]
    probs_1 = probs[:,:,1]

    probs_0 = probs_0.reshape(probs_0.shape[0]*probs_0.shape[1])
    probs_1 = probs_1.reshape(probs_1.shape[0]*probs_1.shape[1])

    assert(probs_0.shape == win_idx.shape)
    assert(probs_1.shape == win_idx.shape)

    pred_mean_0 = calculate_means(indices=win_idx, values=probs_0)
    pred_mean_1 = calculate_means(indices=win_idx, values=probs_1)

    pred_mean_0 = np.expand_dims(pred_mean_0,axis = 1)
    pred_mean_1 = np.expand_dims(pred_mean_1,axis = 1)

    probs_mean = np.concatenate((pred_mean_0,pred_mean_1),axis=1)

    return probs_mean

def calculate_means(indices, values):
    """
    Calculate means of the values that have same index.
    """

    assert(indices.shape == values.shape)

    # Combine indices with predictions
    comb = np.column_stack((indices, values))

    # Sort based on window indices and split when indice changes
    comb = comb[comb[:, 0].argsort()]
    split_on = np.where(np.diff(comb[:, 0]) != 0)[0]+1

    # Take mean from the values that have same index
    #startTime = time.time()
    mean_values = [arr[:, 1].mean() for arr in np.split(comb, split_on)]
    #executionTime = (time.time() - startTime)
    #print('Execution time in seconds: ' + str(executionTime))
    mean_values = np.array(mean_values)

    return mean_values


def challenge_entry(sample_path,model):

    """
    Implementation of AF detection algorithm.
    """

    # Precomputed mean and std from training data.

    #E1
    mean, std =  2.3972106, 2.5281641

    #mean, std = 2.5492988, 2.5239613

    # Some initializations.
    STEP = 256
    FRAME_SIZE = 5120
    stride_sig = 512
    stride_label = int(stride_sig/STEP)

    #Load Signal
    sig, _, fs = load_data(sample_path)
    sig = sig[:, 1]

    sig_len = len(sig)    # Complete signal length. 
    rem_len = sig_len%stride_sig   # Remaining length after making frames.
    div_len = sig_len - rem_len   # Langth used in making frames. 


    tot_labels = int(np.ceil(len(sig)/STEP))

    if div_len < FRAME_SIZE+1:
        pad = np.zeros(STEP-(sig_len%STEP))

        sig_p = np.concatenate((sig,pad))
        data_windows = sig_p.reshape(1,len(sig_p),1)

        # Model Predictions
        probs =model.predict(data_windows, verbose=0)
        predicted = np.argmax(probs[0],axis=1)

        predicted = predicted[:tot_labels]

    else:

        pad = np.zeros(stride_sig-(sig_len%stride_sig))
        sig_p = np.concatenate((sig,pad))

        # Prepare data to pass to model.
        data_windows = prepare_windows(sig_p,win_size = FRAME_SIZE,stride = stride_sig, mean=mean, std=std)

        # prob idx for sliding windows
        n_labels = int(len(sig_p)/STEP)
        probs_win_idx =  get_label_idx(n_labels, stride = stride_label, win_size_label  = 20)

        # Model Predictions
        probs =model.predict(data_windows, verbose=0)

        # Mean probs.
        probs_mean = mean_preds(probs, probs_win_idx)

        # Predictions. 
        predicted = np.argmax(probs_mean,axis=1)
        predicted = predicted[:tot_labels]

    for i in range(3,len(predicted)-3):
        
        if (predicted[i-3:i+4] == [1,1,0,1,0,1,1]).all():
            predicted[i-3:i+4] = 1
        elif (predicted[i-3:i+4] == [0,0,1,0,1,0,0]).all():
            predicted[i-3:i+4] = 0
        elif (predicted[i-2:i+3] == [1,1,0,1,1]).all():
            predicted[i-2:i+3] = 1
        elif (predicted[i-2:i+3] == [0,0,1,0,0]).all():
            predicted[i-2:i+3] = 0


    end_points = []

    # If all labels are 1. with a slight error of +-1 label.  
    if tot_labels -3 < np.sum(predicted) < tot_labels + 2:
        end_points.append([0, len(sig)-1])
        #print("Persisten AF")
    elif not(-1 < np.sum(predicted) < 3):
        #print("Paroxysmal AF")
        state_diff = np.diff(predicted)
        start_r = np.where(state_diff==1)[0] + 1
        end_r = np.where(state_diff==-1)[0] + 1

        if predicted[0] == 1:
            start_r = np.insert(start_r, 0, 0)
        if predicted[-1] == 1:
            end_r = np.insert(end_r, len(end_r), len(predicted)-1)

        start_r = np.expand_dims(start_r, -1)
        end_r = np.expand_dims(end_r, -1)

        start_end = np.concatenate(((start_r*STEP)+128, ((end_r+1)*STEP)-128), axis=-1).tolist()
        end_points.extend(start_end)

        
        if end_points[-1][-1] > len(sig)-1:
            end_points[-1][-1] = len(sig)-1
        if end_points[-1][0] > len(sig)-1:
            end_points[-1][0] = len(sig)-1

    #else:
        #print("Normal")

    pred_dcit = {'predict_endpoints': end_points}


    return pred_dcit


if __name__ == '__main__':
    
    DATA_PATH = sys.argv[1]
    RESULT_PATH = sys.argv[2]

    #DATA_PATH = '../training/'
    #RESULT_PATH = '../results/'

    if not os.path.exists(RESULT_PATH):
        os.makedirs(RESULT_PATH)


    params = {
    "conv_subsample_lengths": [1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2, 1, 2],
    "conv_filter_length": 16,
    "conv_num_filters_start": 32,
    "conv_init": "he_normal",
    "conv_activation": "relu",
    "conv_dropout": 0.2,
    "conv_num_skip": 2,
    "conv_increase_channels_at": 4,
    "learning_rate": 0.001,
    "batch_size": 32,
    "train": "train.json",
    "dev": "dev.json",
    "generator": False,
    "save_dir": "saved",
    "input_shape": [None, 1],
    "num_categories": 2
    }
        
    model = network.build_network(**params)
    model.load_weights("model_entry1_20s.h5")

    test_set = open(os.path.join(DATA_PATH, 'RECORDS'), 'r').read().splitlines()
    for i, sample in enumerate(test_set):
        print(sample)
        #sample = 'data_48_2'
        sample_path = os.path.join(DATA_PATH, sample)

        try: 
            pred_dict = challenge_entry(sample_path,model)
            save_dict(os.path.join(RESULT_PATH, sample+'.json'), pred_dict)
        except: 
            save_dict(os.path.join(RESULT_PATH, sample+'.json'), [])
            print("Exception")




